 ; Program Name: TIMES_4.S
 ;      Version: 1.005

 ; Assembly Instructions:

 ;     Assemble in PC-relative mode and save with a TOS extension.

 ; Execution Instructions:

 ;     Execute from the desktop; or execute SPAWN.TTP, type TIMES_4.TOS on
 ; its command line and view this program's output in TIMES_4.DAT.

 ; Program Function:

 ;     Compares the speed of multiplying by 4 with add dn,dn to asl #2,dn
 ; to determine which is faster.  This experiment attempts to prove or
 ; refute the conclusion advanced by Mike Morton in his magazine article,
 ; "68000 Tricks and Traps", Byte, September, 1986, p163.  As it turns out,
 ; when multiplying a word operand by 4, as Mr. Morton specifies in his
 ; article, the execution times are identical.  But when multiplying a
 ; longword by 4, the shift algorithm is faster.  Furthermore, the shift
 ; algorithm requires less memory.

 ;     Author Stan Kelly-Bootle also asserts that the add.w dn,dn algorithm
 ; is a faster method of multiplying by 4 than is asl #2,dn at the top of
 ; page 202 of his book "680x0 Programming by Example".

calculate_program_size:
 lea        -$102(pc), a1       ; Fetch basepage start address.  
 lea        program_end, a0     ; Fetch program end address.
 trap       #6                  ; Return unused memory to op system.

print_heading:
 lea        heading, a0
 bsr        print_string

word_addition:
 lea        header_1, a0     
 bsr        print_string
 move.l     #49999, d7          ; D7 is counter for the loop.
 trap       #3                  ; Fetch start time.
 move.l     d0, d3              ; Save start_time in D3.
word_addition_loop:             ; Marks start of instruction in the loop.
 add.w      d0, d0              ; To multiply by two.
 add.w      d0, d0              ; To multiply by four.
memory_1:                       ; Marks end of instruction in the loop.
 dbra       d7, word_addition_loop
 trap       #3                  ; Fetch end time.
 sub.l      d3, d0              ; Subtract start time from end time.
 trap       #10                 ; Convert to decimal milliseconds and print.

print_word_addition_requisite_memory:
 lea        header_5, a0
 bsr        print_string
 lea        memory_1, a0           ; Calculate number of bytes occupied by
 lea        word_addition_loop, a1 ; the instructions in the loop.
 suba.l     a1, a0
 bsr        print_requisite_memory

word_shift:
 lea        header_2, a0     
 bsr        print_string
 move.l     #49999, d7          ; D7 is counter for the loop.
 trap       #3                  ; Fetch start time.
 move.l     d0, d3              ; Save start_time in D3.
word_shift_loop:                ; Marks start of instruction in the loop.
 asl.w      #2, d0              ; Shift to multiply by 4.
memory_2:                       ; Marks end of instruction in the loop.
 dbra       d7, word_shift_loop ; Loop 50000 times.
 trap       #3                  ; Fetch end time.
 sub.l      d3, d0              ; Subtract start time from end time.
 trap       #10                 ; Convert to decimal milliseconds and print.

print_word_shift_requisite_memory:
 lea        header_6, a0
 bsr.s      print_string
 lea        memory_2, a0        ; Calculate number of bytes occupied by the
 lea        word_shift_loop, a1 ; instruction in the loop, then store.
 suba.l     a1, a0
 bsr.s      print_requisite_memory

longword_addition:
 lea        header_3, a0     
 bsr.s      print_string
 move.l     #49999, d7          ; D7 is counter for the loop.
 trap       #3                  ; Fetch start time.
 move.l     d0, d3              ; Save start_time in D3.
addition_loop:                  ; Marks start of instructions in the loop.
 add.l      d0, d0              ; To multiply by two.
 add.l      d0, d0              ; To multiply by four.
memory_3:                       ; Marks end of instruction in the loop.
 dbra       d7, addition_loop   ; Loop 50000 times.
 trap       #3                  ; Fetch end time.
 sub.l      d3, d0              ; Subtract start time from end time.
 trap       #10                 ; Convert to decimal milliseconds and print.

print_longword_addition_requisite_memory:
 lea        header_7, a0
 bsr.s      print_string
 lea        memory_3, a0        ; Calculate number of bytes occupied by the
 lea        addition_loop, a1   ; instruction in the loop, then store.
 suba.l     a1, a0
 bsr.s      print_requisite_memory

longword_shift:
 lea        header_4, a0     
 bsr.s      print_string
 move.l     #49999, d7          ; D7 is counter for the loop.
 trap       #3                  ; Fetch start time.
 move.l     d0, d3              ; Save start_time in D3.
 trap       #3                  ; Value of system clock returned in D0.
 move.l     d0, d1              ; Save in trap call register.
shift_loop:                     ; Marks start of instruction in the loop.
 asl.l      #2, d0              ; Shift to multiply by 4.
memory_4:                       ; Marks end of instruction in the loop.
 dbra       d7, shift_loop      ; Loop 50000 times.
 trap       #3                  ; Fetch end time.
 sub.l      d3, d0              ; Subtract start time from end time.
 trap       #10                 ; Convert to decimal milliseconds and print.

print_longword_shift_requisite_memory:
 lea        header_8, a0
 bsr.s      print_string
 lea        memory_4, a0        ; Calculate number of bytes occupied by the
 lea        shift_loop, a1      ; instruction in the loop, then store.
 suba.l     a1, a0
 bsr.s      print_requisite_memory

terminate:
 trap       #8

 ;
 ; SUBROUTINES
 ;

print_requisite_memory:
 move.l     a0, d1
 trap       #4
 bsr.s      print_string
 lea        header_9, a0
 bsr.s      print_string
 rts

print_string:                   ; Expects address of string to be in A0.
 move.l     a0, -(sp)           ; Push address of string onto stack.
 move.w     #9, -(sp)           ; Function = c_conws = GEMDOS $9.
 trap       #1                  ; GEMDOS call
 addq.l     #6, sp              ; Reset stack pointer to top of stack.
 rts

 data
heading:      dc.b  "TIMES_4 Execution Results",$D,$A,$D,$A,0
header_1:     dc.b       "   Word addition time:           ",0
header_2:     dc.b $D,$A,"   Word shift time:              ",0
header_3:     dc.b $D,$A,"   Longword addition time:       ",0
header_4:     dc.b $D,$A,"   Longword shift time:          ",0
header_5:     dc.b       "   Word addition requisite memory:  ",0
header_6:     dc.b       "   Word shift requisite memory:     ",0
header_7:     dc.b       "   Longword add requisite memory:   ",0
header_8:     dc.b       "   Longword shift requisite memory: ",0
header_9:     dc.b       " bytes",$D,$A,0
 bss
 align
program_end:  ds.l   0   
 end
